/*
 * vim: set expandtab tabstop=4 shiftwidth=4 foldmethod=marker:
 *
 * session.c
 *
 * Copyright (C) Keyphrene.com 2005, All rights reserved
 *
 *
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA


 */
#include <Python.h>
#define SSH2_MODULE
#include "ssh2.h"


static char SSH2_Session_setBanner_doc[] = "";

static PyObject *
SSH2_Session_setBanner(SSH2_SessionObj *self, PyObject *args)
{
    char *banner;
    
    if (!PyArg_ParseTuple(args, "s:setBanner", &banner))
        return NULL;
    
    libssh2_banner_set(self->session, banner);
    return PyInt_FromLong(1);
}


static char SSH2_Session_startup_doc[] = "";

static PyObject *
SSH2_Session_startup(SSH2_SessionObj *self, PyObject *args)
{
    PyObject *sock;
    int ret;
    int fd;
    
    if (!PyArg_ParseTuple(args, "O:startup", &sock))
        return NULL;
    
    // Increment the reference count for socket object 
    Py_INCREF(sock);
    self->socket = sock;
    fd = PyObject_AsFileDescriptor(sock);
    
    MY_BEGIN_ALLOW_THREADS(self->tstate);
    ret=libssh2_session_startup(self->session, fd);
    MY_END_ALLOW_THREADS(self->tstate);
    
    if (ret) {
        char *_err = "";
        libssh2_session_last_error(self->session, &_err, NULL, 0);
        PyErr_Format(PySSH2_Error, "SSH startup: %s", _err);
        return NULL;
    }
    self->opened = 1;
    return PyInt_FromLong(1);
}

static char SSH2_Session_close_doc[] = "";

static PyObject *
SSH2_Session_close(SSH2_SessionObj *self, PyObject *args)
{
    char *reason = "end";
    int ret;
    
    if (!PyArg_ParseTuple(args, "|s:disconnect", &reason))
        return NULL;
    
    MY_BEGIN_ALLOW_THREADS(self->tstate);
    ret = libssh2_session_disconnect(self->session, reason);
    MY_END_ALLOW_THREADS(self->tstate);
    
    if(ret) {
        char *_err = "";
        libssh2_session_last_error(self->session, &_err, NULL, 0);
        PyErr_Format(PySSH2_Error, "SSH disconnect: %s", _err);
        return NULL;
    }
    self->opened = 0;
    return PyInt_FromLong(1);
}

static char SSH2_Session_isAuthenticated_doc[] = "";

static PyObject *
SSH2_Session_isAuthenticated(SSH2_SessionObj *self, PyObject *args)
{    
    return PyInt_FromLong(libssh2_userauth_authenticated(self->session)==0?1:0);
}

static char SSH2_Session_getAuthenticationMethods_doc[] = "";

static PyObject *
SSH2_Session_getAuthenticationMethods(SSH2_SessionObj *self, PyObject *args)
{    
    char *user;
    int len=0;
    
    if (!PyArg_ParseTuple(args, "s:getAuthenticationMethods", &user, &len))
        return NULL;
    
    return PyString_FromString(libssh2_userauth_list(self->session, user, len));
}

static char SSH2_Session_getFingerprint_doc[] = "";

static PyObject *
SSH2_Session_getFingerprint(SSH2_SessionObj *self, PyObject *args)
{    
    int hashtype = LIBSSH2_HOSTKEY_HASH_MD5; 
    #define BUFSIZE 21
    char fingerprint[BUFSIZE];
    int i;
    size_t fingerprint_len;
    
    if (!PyArg_ParseTuple(args, "|i:getFingerprint", &hashtype))
        return NULL;

    fingerprint_len = (hashtype == LIBSSH2_HOSTKEY_HASH_MD5) ? 16 : 20;
    
    MY_BEGIN_ALLOW_THREADS(self->tstate);
    const char *hash = libssh2_hostkey_hash(self->session, hashtype);
    MY_END_ALLOW_THREADS(self->tstate);

    for(i = 0; i < fingerprint_len; i++) {
        if (hash[i] == '\0') {
            
            PyErr_SetString(PySSH2_Error, "No fingerprint available using specified hash");
            return NULL;
        }
    
    }
    snprintf(fingerprint, fingerprint_len + 1, "%*s", fingerprint_len, hash);

    return PyString_FromString(fingerprint);
}

static char SSH2_Session_setPassword_doc[] = "";

static PyObject *
SSH2_Session_setPassword(SSH2_SessionObj *self, PyObject *args)
{
    unsigned char *login;
    unsigned char *pwd;
    int ret;
    
    if (!PyArg_ParseTuple(args, "ss:setPassword", &login, &pwd))
        return NULL;
    
    MY_BEGIN_ALLOW_THREADS(self->tstate);
    ret = libssh2_userauth_password(self->session, login, pwd);
    MY_END_ALLOW_THREADS(self->tstate);
    
    if (ret) {
        char *_err = "";
        libssh2_session_last_error(self->session, &_err, NULL, 0);
        PyErr_Format(PySSH2_Error, "Authentication by password failed: %s", _err);
        return NULL;
    }
    return PyInt_FromLong(1);
}

static char SSH2_Session_setPublicKey_doc[] = "";

static PyObject *
SSH2_Session_setPublicKey(SSH2_SessionObj *self, PyObject *args)
{
    char *login;
    char *publickey;
    char *privatekey;
    char *passphrase;
    int ret;
    
    if (!PyArg_ParseTuple(args, "sss|s:setPublicKey", &login, &publickey, &privatekey, &passphrase))
        return NULL;
    
    MY_BEGIN_ALLOW_THREADS(self->tstate);
    ret = libssh2_userauth_publickey_fromfile(self->session, login, publickey, privatekey, passphrase);
    MY_END_ALLOW_THREADS(self->tstate);
    
    if (ret) {
        char *_err = "";
        libssh2_session_last_error(self->session, &_err, NULL, 0);
        PyErr_Format(PySSH2_Error, "Authentication by public key failed: %s", _err);
        return NULL;
    }
    return PyInt_FromLong(1);
}

static char SSH2_Session_getMethods_doc[] = "";

static PyObject *
SSH2_Session_getMethods(SSH2_SessionObj *self, PyObject *args)
{    
    const char *kex, *hostkey, *crypt_cs, *crypt_sc, *mac_cs, *mac_sc, *comp_cs, *comp_sc, *lang_cs, *lang_sc;
    PyObject *methods;
    PyObject *kexp, *hostkeyp, *crypt_csp, *crypt_scp, *mac_csp, *mac_scp, *comp_csp, *comp_scp, *lang_csp, *lang_scp;
    

    kex = libssh2_session_methods(self->session, LIBSSH2_METHOD_KEX);
    hostkey = libssh2_session_methods(self->session, LIBSSH2_METHOD_HOSTKEY);
    crypt_cs = libssh2_session_methods(self->session, LIBSSH2_METHOD_CRYPT_CS);
    crypt_sc = libssh2_session_methods(self->session, LIBSSH2_METHOD_CRYPT_SC);
    mac_cs = libssh2_session_methods(self->session, LIBSSH2_METHOD_MAC_CS);
    mac_sc = libssh2_session_methods(self->session, LIBSSH2_METHOD_MAC_SC);
    comp_cs = libssh2_session_methods(self->session, LIBSSH2_METHOD_COMP_CS);
    comp_sc = libssh2_session_methods(self->session, LIBSSH2_METHOD_COMP_SC);
    lang_cs = libssh2_session_methods(self->session, LIBSSH2_METHOD_LANG_CS);
    lang_sc = libssh2_session_methods(self->session, LIBSSH2_METHOD_LANG_SC);
    
    methods = PyDict_New();

    kexp = (kex == NULL) ? Py_None : PyString_FromString(kex);
    hostkeyp = (hostkey == NULL) ? Py_None : PyString_FromString(hostkey);
    crypt_csp = (crypt_cs == NULL) ? Py_None : PyString_FromString(crypt_cs);
    crypt_scp = (crypt_sc == NULL) ? Py_None : PyString_FromString(crypt_sc);
    mac_csp = (mac_cs == NULL) ? Py_None : PyString_FromString(mac_cs);
    mac_scp = (mac_sc == NULL) ? Py_None : PyString_FromString(mac_sc);
    comp_csp = (comp_cs == NULL) ? Py_None : PyString_FromString(comp_cs);
    comp_scp = (comp_sc == NULL) ? Py_None : PyString_FromString(comp_sc);
    lang_csp = (lang_cs == NULL) ? Py_None : PyString_FromString(lang_cs);
    lang_scp = (lang_sc == NULL) ? Py_None : PyString_FromString(lang_sc);

    PyDict_SetItemString(methods, "KEX", kexp);
    PyDict_SetItemString(methods, "HOSTKEY", hostkeyp);
    PyDict_SetItemString(methods, "CRYPT_CS", crypt_csp);
    PyDict_SetItemString(methods, "CRYPT_SC", crypt_scp);
    PyDict_SetItemString(methods, "MAC_CS", mac_csp);
    PyDict_SetItemString(methods, "MAC_SC", mac_scp);
    PyDict_SetItemString(methods, "COMP_CS", comp_csp);
    PyDict_SetItemString(methods, "COMP_SC", comp_scp);
    PyDict_SetItemString(methods, "LANG_CS", lang_csp);
    PyDict_SetItemString(methods, "LANG_SC", lang_scp);
    
    return methods;
}

static char SSH2_Session_setMethod_doc[] = "";

static PyObject *
SSH2_Session_setMethod(SSH2_SessionObj *self, PyObject *args)
{    
    int method;
    char *pref;
    int _errno;
    _errno = 0;

    if (!PyArg_ParseTuple(args, "is:setMethod", &method, &pref))
        return NULL;
    
    if (libssh2_session_method_pref(self->session, method, pref) == -1) {
        char *_err = "";
        _errno = libssh2_session_last_error(self->session, &_err, NULL, 0);
    }
    

    return PyInt_FromLong(_errno);
}

static char SSH2_Session_Channel_doc[] = "";

static PyObject *
SSH2_Session_Channel(SSH2_SessionObj *self, PyObject *args)
{
    int dealloc = 1;
    
    if (!PyArg_ParseTuple(args, "|i:Channel", &dealloc))
        return NULL;

    return (PyObject *)SSH2_Channel_New(libssh2_channel_open_session(self->session), dealloc);
}

static char SSH2_Session_SCPGet_doc[] = "";

static PyObject *
SSH2_Session_SCPGet(SSH2_SessionObj *self, PyObject *args)
{
    char *path;
    LIBSSH2_CHANNEL *channel;
    //struct stat {
    //    dev_t     st_dev;      /* Périphérique                  */
    //    ino_t     st_ino;      /* Numéro i-nœud             */
    //    mode_t    st_mode;     /* Protection                    */
    //    nlink_t   st_nlink;    /* Nb liens matériels            */
    //    uid_t     st_uid;      /* UID propriétaire              */
    //    gid_t     st_gid;      /* GID propriétaire              */
    //    dev_t     st_rdev;     /* Type périphérique             */
    //    off_t     st_size;     /* Taille totale en octets       */
    //    blksize_t st_blksize;  /* Taille de bloc pour E/S       */
    //    blkcnt_t  st_blocks;   /* Nombre de blocs alloués       */
    //    time_t    st_atime;    /* Heure dernier accès           */
    //    time_t    st_mtime;    /* Heure dernière modification   */
    //    time_t    st_ctime;    /* Heure dernier changement état */

    //} sb;
    struct stat sb;

    PyObject *dstfile;
    PyObject *desc;

    if (!PyArg_ParseTuple(args, "s:SCPGet", &path))
        return NULL;

    // channel = libssh2_scp_recv(self->session, path, NULL); // &sb
    channel = libssh2_scp_recv(self->session, path, &sb);
    if (channel == NULL) {
        char *_err = "";
        libssh2_session_last_error(self->session, &_err, NULL, 0);
        PyErr_Format(PySSH2_Error, "SCP error: %s", _err);
        return NULL;
    }

    dstfile = PyList_New(2);
    PyList_SET_ITEM(dstfile,0,(PyObject *)SSH2_Channel_New(channel, 1));

    desc = PyList_New(4);
    PyList_SET_ITEM(desc,0,PyInt_FromLong(sb.st_size));
    PyList_SET_ITEM(desc,1,PyInt_FromLong(sb.st_mtime));
    PyList_SET_ITEM(desc,2,PyInt_FromLong(sb.st_atime));
    PyList_SET_ITEM(desc,3,PyInt_FromLong(sb.st_mode));

    PyList_SET_ITEM(dstfile,1,desc);
    return (PyObject *)dstfile;
}

static char SSH2_Session_SCPPut_doc[] = "";

static PyObject *
SSH2_Session_SCPPut(SSH2_SessionObj *self, PyObject *args)
{
    char *path;
    int mode;
    unsigned long filesize;
    LIBSSH2_CHANNEL *channel;
    
    if (!PyArg_ParseTuple(args, "sik:SCPPut", &path, &mode, &filesize))
        return NULL;

    channel = libssh2_scp_send(self->session, path, mode, filesize);
    if (channel == NULL) {
        char *_err = "";
        libssh2_session_last_error(self->session, &_err, NULL, 0);
        PyErr_Format(PySSH2_Error, "SCP error: %s", _err);
        return NULL;
    }
    return (PyObject *)SSH2_Channel_New(channel, 1);
}

static char SSH2_Session_SFTP_doc[] = "";

static PyObject *
SSH2_Session_SFTP(SSH2_SessionObj *self, PyObject *args)
{
    int dealloc = 1;
    
    if (!PyArg_ParseTuple(args, "|i:SFTP", &dealloc))
        return NULL;

    return (PyObject *)SSH2_SFTP_New(libssh2_sftp_init(self->session), dealloc);
}



static char SSH2_Session_DirectTcpIP_doc[] = "";

static PyObject *
SSH2_Session_DirectTcpIP(SSH2_SessionObj *self, PyObject *args)
{
    char *host;
    char *shost = "127.0.0.1";
    int port;
    int sport = 22;
    LIBSSH2_CHANNEL *channel;
    
    if (!PyArg_ParseTuple(args, "si|si:DirectTcpIP", &host, &port, &shost, &sport))
        return NULL;
    
    if (self->debug) {
        printf("DirectTcpIp %s:%d %s:%d\n", host, port, shost, sport);
    }
    MY_BEGIN_ALLOW_THREADS(self->tstate);
    channel = libssh2_channel_direct_tcpip_ex(self->session, host, port, shost, sport);
    //~ channel = libssh2_channel_direct_tcpip(self->session, host,port);
    MY_END_ALLOW_THREADS(self->tstate);
    
    if (channel == NULL) {
        char *_err = "";
        libssh2_session_last_error(self->session, &_err, NULL, 0);
        PyErr_Format(PySSH2_Error, "DirectTcpIP error: %s", _err);
        return NULL;
    }
    return (PyObject *)channel;
}

static char SSH2_Session_ForwardListen_doc[] = "";
// host: "0.0.0.0"
// port to bind
static PyObject *
SSH2_Session_ForwardListen(SSH2_SessionObj *self, PyObject *args)
{
    char *host;
    int port;
    int queue_maxsize;
    int *bound_port;
    LIBSSH2_LISTENER *listener;
    
    if (!PyArg_ParseTuple(args, "siii:ForwardListen", &host, &port, &bound_port, &queue_maxsize))
        return NULL;

    MY_BEGIN_ALLOW_THREADS(self->tstate);
    listener = libssh2_channel_forward_listen_ex(self->session, host, port, bound_port, queue_maxsize);
    MY_END_ALLOW_THREADS(self->tstate);
    
    if (listener == NULL) {
        char *_err = "";
        libssh2_session_last_error(self->session, &_err, NULL, 0);
        PyErr_Format(PySSH2_Error, "ForwardListen error: %s", _err);
        return NULL;
    }
    return (PyObject *)SSH2_Listener_New(listener, 0);
}

static char SSH2_Session_popen_doc[] = "";
static PyObject *
SSH2_Session_popen(SSH2_SessionObj *self, PyObject *args)
{
    LIBSSH2_CHANNEL *channel;
    char *command;

    if (!PyArg_ParseTuple(args, "s:popen", &command))
        return NULL;

    channel = libssh2_channel_open_session(self->session);
    if (channel == NULL) {
        PyErr_SetString(PySSH2_Error, "SSH error.");
        return NULL;
    }
    if (libssh2_channel_exec(channel, command)) {
        PyErr_SetString(PySSH2_Error, "Unable to request command execution on remote host");
        libssh2_channel_free(channel);
    }

     return (PyObject *)(SSH2_Channel_New(channel, 1)); 
}

static char SSH2_Session_setDebug_doc[] = "";
static PyObject *
SSH2_Session_setDebug(SSH2_SessionObj *self, PyObject *args)
{
     self->debug = 1;
     return PyInt_FromLong(1); 
}


/*
 *
 * Error codes
 *
 */
static char SSH2_Session_bindErrors_doc[] = "\n\
return dict(errno,\"errtype\"...)\n\
";
static PyObject *
SSH2_Session_bindErrors(PyObject * args)
{
return Py_BuildValue( "{i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s, i:s}",
    LIBSSH2_ERROR_ALLOC, "LIBSSH2_ERROR_ALLOC",
    LIBSSH2_ERROR_SOCKET_NONE, "LIBSSH2_ERROR_SOCKET_NONE",
    LIBSSH2_ERROR_BANNER_NONE, "LIBSSH2_ERROR_BANNER_NONE",
    LIBSSH2_ERROR_BANNER_SEND, "LIBSSH2_ERROR_BANNER_SEND",
    LIBSSH2_ERROR_INVALID_MAC, "LIBSSH2_ERROR_INVALID_MAC",
    LIBSSH2_ERROR_KEX_FAILURE, "LIBSSH2_ERROR_KEX_FAILURE",
    LIBSSH2_ERROR_SOCKET_SEND, "LIBSSH2_ERROR_SOCKET_SEND",
    LIBSSH2_ERROR_KEY_EXCHANGE_FAILURE, "LIBSSH2_ERROR_KEY_EXCHANGE_FAILURE",
    LIBSSH2_ERROR_TIMEOUT, "LIBSSH2_ERROR_TIMEOUT",
    LIBSSH2_ERROR_HOSTKEY_INIT, "LIBSSH2_ERROR_HOSTKEY_INIT",
    LIBSSH2_ERROR_HOSTKEY_SIGN, "LIBSSH2_ERROR_HOSTKEY_SIGN",
    LIBSSH2_ERROR_DECRYPT, "LIBSSH2_ERROR_DECRYPT",
    LIBSSH2_ERROR_SOCKET_DISCONNECT, "LIBSSH2_ERROR_SOCKET_DISCONNECT",
    LIBSSH2_ERROR_PROTO, "LIBSSH2_ERROR_PROTO",
    LIBSSH2_ERROR_PASSWORD_EXPIRED, "LIBSSH2_ERROR_PASSWORD_EXPIRED",
    LIBSSH2_ERROR_FILE, "LIBSSH2_ERROR_FILE",
    LIBSSH2_ERROR_METHOD_NONE, "LIBSSH2_ERROR_METHOD_NONE",
    LIBSSH2_ERROR_PUBLICKEY_UNRECOGNIZED, "LIBSSH2_ERROR_PUBLICKEY_UNRECOGNIZED",
    LIBSSH2_ERROR_PUBLICKEY_UNVERIFIED, "LIBSSH2_ERROR_PUBLICKEY_UNVERIFIED",
    LIBSSH2_ERROR_CHANNEL_OUTOFORDER, "LIBSSH2_ERROR_CHANNEL_OUTOFORDER",
    LIBSSH2_ERROR_CHANNEL_FAILURE, "LIBSSH2_ERROR_CHANNEL_FAILURE",
    LIBSSH2_ERROR_CHANNEL_REQUEST_DENIED, "LIBSSH2_ERROR_CHANNEL_REQUEST_DENIED",
    LIBSSH2_ERROR_CHANNEL_UNKNOWN, "LIBSSH2_ERROR_CHANNEL_UNKNOWN",
    LIBSSH2_ERROR_CHANNEL_WINDOW_EXCEEDED, "LIBSSH2_ERROR_CHANNEL_WINDOW_EXCEEDED",
    LIBSSH2_ERROR_CHANNEL_PACKET_EXCEEDED, "LIBSSH2_ERROR_CHANNEL_PACKET_EXCEEDED",
    LIBSSH2_ERROR_CHANNEL_CLOSED, "LIBSSH2_ERROR_CHANNEL_CLOSED",
    LIBSSH2_ERROR_CHANNEL_EOF_SENT, "LIBSSH2_ERROR_CHANNEL_EOF_SENT",
    LIBSSH2_ERROR_SCP_PROTOCOL, "LIBSSH2_ERROR_SCP_PROTOCOL",
    LIBSSH2_ERROR_ZLIB, "LIBSSH2_ERROR_ZLIB",
    LIBSSH2_ERROR_SOCKET_TIMEOUT, "LIBSSH2_ERROR_SOCKET_TIMEOUT",
    LIBSSH2_ERROR_REQUEST_DENIED, "LIBSSH2_ERROR_REQUEST_DENIED",
    LIBSSH2_ERROR_METHOD_NOT_SUPPORTED, "LIBSSH2_ERROR_METHOD_NOT_SUPPORTED",
    LIBSSH2_ERROR_INVAL, "LIBSSH2_ERROR_INVAL",
    LIBSSH2_ERROR_SFTP_PROTOCOL, "LIBSSH2_ERROR_SFTP_PROTOCOL",
    LIBSSH2_FX_OK, "LIBSSH2_FX_OK",
    LIBSSH2_FX_EOF, "LIBSSH2_FX_EOF",
    LIBSSH2_FX_NO_SUCH_FILE, "LIBSSH2_FX_NO_SUCH_FILE",
    LIBSSH2_FX_PERMISSION_DENIED, "LIBSSH2_FX_PERMISSION_DENIED",
    LIBSSH2_FX_FAILURE, "LIBSSH2_FX_FAILURE",
    LIBSSH2_FX_BAD_MESSAGE, "LIBSSH2_FX_BAD_MESSAGE",
    LIBSSH2_FX_NO_CONNECTION, "LIBSSH2_FX_NO_CONNECTION",
    LIBSSH2_FX_CONNECTION_LOST, "LIBSSH2_FX_CONNECTION_LOST",
    LIBSSH2_FX_OP_UNSUPPORTED, "LIBSSH2_FX_OP_UNSUPPORTED",
    LIBSSH2_FX_INVALID_HANDLE, "LIBSSH2_FX_INVALID_HANDLE",
    LIBSSH2_FX_NO_SUCH_PATH, "LIBSSH2_FX_NO_SUCH_PATH",
    LIBSSH2_FX_FILE_ALREADY_EXISTS, "LIBSSH2_FX_FILE_ALREADY_EXISTS",
    LIBSSH2_FX_WRITE_PROTECT, "LIBSSH2_FX_WRITE_PROTECT",
    LIBSSH2_FX_NO_MEDIA, "LIBSSH2_FX_NO_MEDIA",
    LIBSSH2_FX_NO_SPACE_ON_FILESYSTEM, "LIBSSH2_FX_NO_SPACE_ON_FILESYSTEM",
    LIBSSH2_FX_QUOTA_EXCEEDED, "LIBSSH2_FX_QUOTA_EXCEEDED",
    LIBSSH2_FX_UNKNOWN_PRINCIPLE, "LIBSSH2_FX_UNKNOWN_PRINCIPLE",
    /* LIBSSH2_FX_LOCK_CONF, "LIBSSH2_FX_LOCK_CONF", */
    LIBSSH2_FX_DIR_NOT_EMPTY, "LIBSSH2_FX_DIR_NOT_EMPTY",
    LIBSSH2_FX_NOT_A_DIRECTORY, "LIBSSH2_FX_NOT_A_DIRECTORY",
    LIBSSH2_FX_INVALID_FILENAME, "LIBSSH2_FX_INVALID_FILENAME",
    LIBSSH2_FX_LINK_LOOP, "LIBSSH2_FX_LINK_LOOP");

}

/*
 * ADD_METHOD(name) expands to a correct PyMethodDef declaration
 *   {  'name', (PyCFunction)crypto_Session_name, METH_VARARGS }
 * for convenience
 */
#define ADD_METHOD(name) { #name, (PyCFunction)SSH2_Session_##name, METH_VARARGS, SSH2_Session_##name##_doc }
static PyMethodDef SSH2_Session_methods[] =
{
    ADD_METHOD(setBanner),
    ADD_METHOD(startup),
    ADD_METHOD(close),
    ADD_METHOD(isAuthenticated),
    ADD_METHOD(getFingerprint),
    ADD_METHOD(getAuthenticationMethods),
    ADD_METHOD(setPassword),
    ADD_METHOD(setPublicKey),
    ADD_METHOD(getMethods),
    ADD_METHOD(setMethod),
    ADD_METHOD(Channel),
    ADD_METHOD(SFTP),
    ADD_METHOD(SCPGet),
    ADD_METHOD(SCPPut),
    ADD_METHOD(DirectTcpIP),
    ADD_METHOD(ForwardListen),
    ADD_METHOD(popen),
    ADD_METHOD(setDebug),
    ADD_METHOD(bindErrors),
    { NULL, NULL }
};
#undef ADD_METHOD


/*
 * Constructor for Session objects, never called by Python code directly
 *
 * Arguments: cert    - A "real" Session certificate object
 *            dealloc - Boolean value to specify whether the destructor should
 *                      free the "real" Session object
 * Returns:   The newly created Session object
 */
SSH2_SessionObj *
SSH2_Session_New(LIBSSH2_SESSION *session, int dealloc)
{
    SSH2_SessionObj *self;

    self = PyObject_New(SSH2_SessionObj, &SSH2_Session_Type);

    if (self == NULL)
        return NULL;

    self->session = session;
    self->dealloc = dealloc;
    self->opened = 0;
    self->debug = 0;
    self->socket=NULL;
    self->tstate = NULL;
    
    libssh2_banner_set(session, LIBSSH2_SSH_DEFAULT_BANNER " Python");

    return self;
}

/*
 * Deallocate the memory used by the Session object
 *
 * Arguments: self - The Session object
 * Returns:   None
 */
/* static void */
static void
SSH2_Session_dealloc(SSH2_SessionObj *self)
{
    /* Sometimes we don't have to dealloc the "real" Session pointer ourselves */
    char strdebug[200] = "";

    if (self->opened) {
        strcat(strdebug,"Close SSH2 Session\n");
        libssh2_session_disconnect(self->session, "end");
    }
    if (self->dealloc) {
        strcat(strdebug,"Dealloc SSH2 Session\n");
        libssh2_session_free(self->session);
    }
    
    Py_XDECREF(self->socket);
    self->socket = NULL;
    strcat(strdebug,"Destroy SSH2 Session\n");
    strcat(strdebug,"Destroy SSH2 TUT Session\n");
    if (self) {
        strcat(strdebug,"Destroy SSH3 Session\n");
        PyObject_Del(self);
    }
    strcat(strdebug,"Destroy Totally\n");

    if (self->debug) {
        printf(strdebug);
    }
    // free(strdebug);
}

/*
 * Find attribute
 *
 * Arguments: self - The Session object
 *            name - The attribute name
 * Returns:   A Python object for the attribute, or NULL if something went
 *            wrong
 */
static PyObject *
SSH2_Session_getattr(SSH2_SessionObj *self, char *name)
{
    return Py_FindMethod(SSH2_Session_methods, (PyObject *)self, name);
}

PyTypeObject SSH2_Session_Type = {
    PyObject_HEAD_INIT(NULL)
    0,
    "Session",
    sizeof(SSH2_SessionObj),
    0,
    (destructor)SSH2_Session_dealloc,
    NULL, /* print */
    (getattrfunc)SSH2_Session_getattr,
    NULL, /* setattr */
    NULL, /* compare */
    NULL, /* repr */
    NULL, /* as_number */
    NULL, /* as_sequence */
    NULL, /* as_mapping */
    NULL, /* hash */
};

/*
 * Initialize the Session
 *
 * Arguments: dict - The SSH2 module dictionary
 * Returns:   None
 */
int
init_SSH2_Session(PyObject *dict)
{
    SSH2_Session_Type.ob_type = &PyType_Type;
    Py_INCREF(&SSH2_Session_Type);
    PyDict_SetItemString(dict, "SessionType", (PyObject *)&SSH2_Session_Type);
    return 1;
}


